/*             This file is part of the New World OS project
--                 Copyright (C) 2004-2007  QRW Software
--           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--                      http://www.qrwsoftware.com
--                      http://nwos.sourceforge.com
--
-- NWOS is free software;  you can redistribute it and/or modify it under the
-- terms of the GNU General Public License  as published by the Free Software
-- Foundation; either version 2, or (at your option) any later version.  This
-- software is distributed with the hope that it will be useful,  but WITHOUT
-- ANY WARRANTY;  without  even the  implied warranty  of MERCHANTABILITY  or
-- FITNESS FOR A PARTICULAR PURPOSE.   See the GNU General Public License for
-- more  details.  You should have received a copy  of the GNU General Public
-- License along with this package;  see the file LICENSE.  If not, write to:
--
--      Free Software Foundation, Inc.
--      59 Temple Place - Suite 330
--      Boston, MA 02111-1307, USA.
--
-- $Log: reference_list.c,v $
-- Revision 1.16  2007/06/30 14:14:01  jsedwards
-- Added #ifdef around logging of add_to_reference_list, so it is off by default.
--
-- Revision 1.15  2007/05/28 16:03:25  jsedwards
-- Added 'remove_from_references' and 'remove_from_reference_list' functions.
--
-- Revision 1.14  2007/05/28 09:06:48  jsedwards
-- Fixed the comment added in previous revision.  Still NO code changes.
--
-- Revision 1.13  2007/05/28 09:04:53  jsedwards
-- Added comment, NO code changes.
--
-- Revision 1.12  2007/02/11 14:32:39  jsedwards
-- Changed calls to 'sprintf' to 'snprintf' so OpenBSD won't whine so much.
--
-- Revision 1.11  2007/01/09 03:02:34  jsedwards
-- Fixed bug where if there were exactly 54 entries in the reference list it
-- wrote a random value into the next block pointer (which has to be null in
-- this case).
--
-- Revision 1.10  2007/01/05 13:44:15  jsedwards
-- Added asserts to verify the reference_list is being created in the correct
-- space (public/private).
--
-- Revision 1.9  2006/12/29 01:49:45  jsedwards
-- Fix compile error.
--
-- Revision 1.8  2006/12/28 23:17:23  jsedwards
-- Removed line feed from messages to go into the log.
--
-- Revision 1.7  2006/12/25 12:18:34  jsedwards
-- Fix error in printing reference to log in add_reference.
--
-- Revision 1.6  2006/12/20 13:12:48  jsedwards
-- Added malloc_class_reference_list function (untested).
--
-- Revision 1.5  2006/12/15 15:01:38  jsedwards
-- Changed for only two disk areas (Public and Private-Encrypted).
--
-- Revision 1.4  2006/12/05 13:56:33  jsedwards
-- Fix so that it doesn't fill unencrypted blocks with random data.
--
-- Revision 1.3  2006/12/04 04:42:05  jsedwards
-- Changed to call new generate_new_id_of_same_type function.
--
-- Revision 1.2  2006/12/01 14:24:31  jsedwards
-- Added malloc and free reference list functions.
--
-- Revision 1.1  2006/11/11 15:21:40  jsedwards
-- Moved all reference list code out of objectify.c and into this file.
--
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "crc32.h"
#include "objectify_private.h"



#define REF_LIST_CACHE_SIZE 64

typedef struct {
  ObjRef ref;
  int num_refs;
  Ref_List_First_Block* first_block;
  int age;
  uint8 ivec[IVEC_SIZE];
} Block_Cache;


static int ref_list_tick = 0;

static Block_Cache ref_list_cache[REF_LIST_CACHE_SIZE];


static void check_ref_list_data(ObjRef* ref, uint8* data, size_t data_size, uint8 stored[4])
{
    uint8 computed[4];

#if 0
    printf("Total size: %d\n", total_size);
    printf("ObjectHeader: %d\n", sizeof(EveryObject));
    printf("object pointer: %p\n", object);
    {
      int i;
      uint8* ptr = (uint8*)object + sizeof(EveryObject);
      printf("data pointer: %p\n", ptr);
      printf("data: ");
      for (i = 0; i < data_size; i++) printf("%02x%c", *ptr++, i%16==15?'\n':' ');
      printf("\n");
    }
#endif

    nwos_crc32_calculate((uint8*)data, data_size, computed);

    if (stored[0] != computed[0] || stored[1] != computed[1] || stored[2] != computed[2] || stored[3] != computed[3])
    {
	printf("reference list: %02x%02x%02x%02x - bad data checksum, ", ref->id[0], ref->id[1], ref->id[2], ref->id[3]);
	printf("size %zd - ", data_size);
	printf("computed %02x%02x%02x%02x - ", computed[0], computed[1], computed[2], computed[3]);
	printf("stored %02x%02x%02x%02x\n", stored[0], stored[1], stored[2], stored[3]);

	printf("\n");
	{
	  int i;
	  for (i = 0; i < data_size; i++)
	  {
	      printf("%02x%c", data[i], (i % 16) == 15 ? '\n' : ' ');
	  }
	  printf("\n");
	}
	exit(1);
    }
}


static void read_ref_list_into_cache(Block_Cache* cache)
{
    uint8 ivec[IVEC_SIZE];
    Ref_List_Extra_Block* prev_ptr;            /* points to the previous block */
    Ref_List_Extra_Block* next_ptr;            /* points to the next block */
    int i;
    int seq;

    assert(sizeof(Ref_List_First_Block) == FILE_BLOCK_SIZE + sizeof(void*));
    assert(sizeof(Ref_List_Extra_Block) == FILE_BLOCK_SIZE + sizeof(void*));

    memset(ivec, 0, sizeof(ivec));

    cache->first_block = nwos_malloc(sizeof(Ref_List_First_Block));
    assert(cache->first_block != NULL);

    memset(ivec, 0, sizeof(ivec));
    memcpy(cache->ivec, ivec, sizeof(cache->ivec));

    cache->first_block->next_block_ptr = NULL;

    nwos_read_object_from_disk_and_decrypt(&cache->ref, &cache->first_block->list, FILE_BLOCK_SIZE, ivec, nwos_random_sequence[0]);

    assert(is_same_object(&cache->first_block->list.common_header.class_definition, &nwos_reference_list_class_ref));

    cache->num_refs = 0;

    while (cache->num_refs < MAX_REFS_IN_REF_LIST)
    {
	if (is_void_reference(&cache->first_block->refs[cache->num_refs])) break;
	cache->num_refs++;
    }

    /* use the flags as a pointer to the next block */
    prev_ptr = (Ref_List_Extra_Block*) cache->first_block;

    check_ref_list_data(&cache->ref,
			(uint8*) cache->first_block->refs,
			(cache->num_refs + 1) * sizeof(ObjRef),                 /* include void or next_block_ref */
			cache->first_block->list.common_header.data_chksum);

    if (cache->num_refs < MAX_REFS_IN_REF_LIST)
    {
	prev_ptr->next_block_ref.word = 0;
    }

    assert(prev_ptr->next_block_ptr == NULL);

    seq = 1;    /* use different sequence tables for each block in turn */

    while (!is_void_reference(&prev_ptr->next_block_ref))    /* more blocks in this list */
    {
	next_ptr = nwos_malloc(sizeof(Ref_List_Extra_Block));

	assert(next_ptr != NULL);

	next_ptr->next_block_ptr = NULL;

	memcpy(cache->ivec, ivec, sizeof(cache->ivec));  /* save this ivec in case this block has to be written back to disk */

	nwos_read_object_from_disk_and_decrypt(&prev_ptr->next_block_ref, &next_ptr->dirty, FILE_BLOCK_SIZE, ivec, nwos_random_sequence[seq]);

	for (i = 0; i < MAX_REFS_IN_SIDECAR; i++)
	{
	    if (is_void_reference(&next_ptr->refs[i])) break;
	    cache->num_refs++;
	}

	check_ref_list_data(&cache->ref, (uint8*) next_ptr->refs, (i + 1) * sizeof(ObjRef), next_ptr->checksum);

	if (i < MAX_REFS_IN_SIDECAR)    /* the link will be garbage, make it void */
	{
	    next_ptr->next_block_ref.word = 0;
	}
	else
	{
	    assert(!is_same_object(&prev_ptr->next_block_ref, &next_ptr->next_block_ref));
	}

	prev_ptr->next_block_ptr = next_ptr;

	prev_ptr = next_ptr;

	seq = (seq + 1) % NUM_STORED_SEQ;    /* wrap around if we've used all of them */
    }
}



static void write_reference_list(Block_Cache* cache)
{
    uint8 ivec[IVEC_SIZE];
    Ref_List_Extra_Block *block = NULL;
    int running_total;
    int refs_remaining;
    Ref_List_Extra_Block* prev_ptr = NULL;
    Ref_List_Extra_Block* next_ptr = NULL;
    ObjRef ref;
    int i;
    int seq;
    bool encrypted;

    encrypted = (nwos_reference_type(&cache->ref) == Private_Reference);

    memset(ivec, 0, sizeof(ivec));

    if (cache->num_refs <= MAX_REFS_IN_REF_LIST)    /* there is just one block */
    {
	assert(cache->first_block->next_block_ptr == NULL);  /* the next block pointer should be null */
	assert(is_void_reference(&cache->first_block->refs[cache->num_refs]));

	if (cache->first_block->list.common_header.flags == 0xffffffff)    /* block is dirty, needs to be written to disk */
	{
	    for (i = cache->num_refs + 1; i < MAX_REFS_IN_REF_LIST; i++)
	    {
		if (encrypted)
		{
		    cache->first_block->refs[i].word = (random() << 1) ^ (random() >> 1);
		}
		else
		{
		    cache->first_block->refs[i].word = 0;
		}
	    }

	    if (encrypted && cache->num_refs < MAX_REFS_IN_REF_LIST)
	    {
		cache->first_block->next_block_ref.word = (random() << 1) ^ (random() >> 1);
	    }
	    else
	    {
		cache->first_block->next_block_ref.word = 0;
	    }

	    nwos_crc32_calculate((uint8*) &cache->first_block->refs, 
				 (cache->num_refs + 1) * sizeof(ObjRef),               /* include void ptr or next_block_ref */
				 cache->first_block->list.common_header.data_chksum);

	    block = (Ref_List_Extra_Block*)cache->first_block;     /* flag that it needs to be written */

	    seq = 0;
	}
	else                    /* we don't need to write this block so we can get rid of it now */
	{
	    nwos_free(cache->first_block);
	}
    }
    else
    {
	assert(cache->first_block->next_block_ptr != NULL);  /* the next block pointer should not be null */

	running_total = MAX_REFS_IN_REF_LIST;
	prev_ptr = (Ref_List_Extra_Block*)cache->first_block;

	seq = 0;

	while (running_total < cache->num_refs)    /* read more blocks */
	{
	    next_ptr = prev_ptr->next_block_ptr;

	    assert(is_same_object(&prev_ptr->next_block_ref, &next_ptr->id));  /* next pointer ref matches */

	    nwos_free(prev_ptr);   /* won't be needing this anymore */

	    refs_remaining = cache->num_refs - running_total;

	    if (refs_remaining <= MAX_REFS_IN_SIDECAR)
	    {
		assert(next_ptr->next_block_ptr == NULL);  /* the next block pointer should be null */
		assert(is_void_reference(&next_ptr->refs[refs_remaining]));

		if (next_ptr->dirty == -1)    /* block is dirty, needs to be written to disk */
		{

		    for (i = refs_remaining + 1; i < MAX_REFS_IN_SIDECAR; i++)
		    {
			if (encrypted)
			{
			    next_ptr->refs[i].word = (random() << 1) ^ (random() >> 1);
			}
			else
			{
			    next_ptr->refs[i].word = 0;
			}
		    }

		    if (encrypted && refs_remaining < MAX_REFS_IN_SIDECAR)
		    {
			next_ptr->next_block_ref.word = (random() << 1) ^ (random() >> 1);    /* fill next with random junk */
		    }
		    else
		    {
			next_ptr->next_block_ref.word = 0;    /* mark the end */
		    }

		    nwos_crc32_calculate((uint8*) &next_ptr->refs, 
					 (refs_remaining + 1) * sizeof(ObjRef),        /* include void ptr or next_block_ref */
					 next_ptr->checksum);

		    block = next_ptr;     /* flag that it needs to be written */
		}

		running_total += refs_remaining;
	    }
	    else
	    {
		assert(next_ptr->next_block_ptr != NULL); /* the next block pointer should not be null */
		running_total += MAX_REFS_IN_SIDECAR;
	    }
	    
	    prev_ptr = next_ptr;

	    seq = (seq + 1) % NUM_STORED_SEQ;    /* wrap around if we've used all of them */
	}


	if (block == NULL)      /* we don't need to write this block so we can get rid of it now */
	{
	    assert(next_ptr != NULL);

	    nwos_free(next_ptr);
	}

	assert(running_total == cache->num_refs);
    }

    if (block != NULL)    /* it needs to be written */
    {
	assert(block->dirty == 0xffffffff);

	/* write object expects these to be null right now */
	block->dirty = 0;

	copy_reference(&ref, &block->id);

	/* printf("writing block: %02x%02x%02x%02x\n", ref.id[0], ref.id[1], ref.id[2], ref.id[3]); */

	nwos_write_object_to_disk_and_encrypt(&ref, &block->dirty, FILE_BLOCK_SIZE, cache->ivec, nwos_random_sequence[seq]);

	nwos_free(block);
    }

    /* just in case, nuke these */
    cache->first_block = NULL;
    memset(&cache->ref, 0, sizeof(cache->ref));
}


static Block_Cache* find_ref_list_in_cache(ObjRef* ref)
{
    int i;
    Block_Cache* result;

    assert(!is_void_reference(ref));

    for (i = 0; i < REF_LIST_CACHE_SIZE; i++)
    {
	if (is_same_object(&ref_list_cache[i].ref, ref)) break;
    }

    if (i < REF_LIST_CACHE_SIZE)   /* found it */
    {
	result = &ref_list_cache[i];
    }
    else                           /* didn't find it */
    {
	/* find an empty one or find the oldest */

	result = ref_list_cache;

	for (i = 0; i < REF_LIST_CACHE_SIZE; i++)
	{
	    if (is_void_reference(&ref_list_cache[i].ref))
	    {
		result = &ref_list_cache[i];
		break;
	    }

	    if (ref_list_cache[i].age < result->age)
	    {
		result = &ref_list_cache[i];
	    }
	}

	if (i == REF_LIST_CACHE_SIZE)    /* didn't find an empty one, write the oldest one out */
	{
	    write_reference_list(result);
	}

	memcpy(&result->ref, ref, sizeof(result->ref));

	read_ref_list_into_cache(result);
    }

    ref_list_tick++;
    result->age = ref_list_tick;

    return result;
}


/* WARNING: this only works on reference lists */

size_t nwos_reference_list_size(ObjRef* ref)
{
    Block_Cache* cache;

    assert(!is_void_reference(ref));

    cache = find_ref_list_in_cache(ref);

    return sizeof(ReferenceList) + (cache->num_refs * sizeof(ObjRef));
}


void nwos_read_reference_list_from_disk(ObjRef* ref, ReferenceList* object, size_t size)
{
    Block_Cache* cache;
    int running_total;
    int refs_remaining;
    Ref_List_Extra_Block* next_ptr;

    assert(!is_void_reference(ref));

    cache = find_ref_list_in_cache(ref);

    assert(size >= sizeof(ReferenceList) + cache->num_refs * sizeof(ObjRef));

    /* copy the first block (which is a reference list object */

    if (cache->num_refs <= MAX_REFS_IN_REF_LIST)    /* there is just one block */
    {
	assert(cache->first_block->next_block_ptr == NULL);         /* the next block pointer should be null */
	assert(is_void_reference(&cache->first_block->refs[cache->num_refs])); /* this should always be null */

	memcpy(object, &cache->first_block->list, sizeof(ReferenceList) + cache->num_refs * sizeof(ObjRef));
    }
    else
    {
	assert(cache->first_block->next_block_ptr != NULL);         /* the next block pointer should not be null */
	assert(sizeof(ReferenceList) + MAX_REFS_IN_REF_LIST * sizeof(ObjRef) == FILE_BLOCK_SIZE - 4);

	memcpy(object, &cache->first_block->list, sizeof(ReferenceList) + MAX_REFS_IN_REF_LIST * sizeof(ObjRef));

	running_total = MAX_REFS_IN_REF_LIST;
	next_ptr = *(Ref_List_Extra_Block**)cache->first_block;        /* get pointer to first sidecar */

	assert(is_same_object(&cache->first_block->next_block_ref, &next_ptr->id));  /* next pointer ref matches */

	while (running_total < cache->num_refs)    /* read more blocks */
	{
	    refs_remaining = cache->num_refs - running_total;

	    if (refs_remaining <= MAX_REFS_IN_SIDECAR)
	    {
		assert(next_ptr->next_block_ptr == NULL);  /* the next block pointer should be null */
		assert(is_void_reference(&next_ptr->refs[refs_remaining]));
		memcpy(&object->references[running_total], next_ptr->refs, refs_remaining * sizeof(ObjRef));
		running_total += refs_remaining;
	    }
	    else
	    {
		assert(next_ptr->next_block_ptr != NULL);  /* the next block pointer should not be null */
		memcpy(&object->references[running_total], next_ptr->refs, MAX_REFS_IN_SIDECAR * sizeof(ObjRef));
		running_total += MAX_REFS_IN_SIDECAR;

		/* next pointer ref matches */
		assert(is_same_object(&next_ptr->next_block_ref, &next_ptr->next_block_ptr->id));
	    }

	    next_ptr = next_ptr->next_block_ptr;
	}
	assert(running_total == cache->num_refs);
    }
}


void nwos_create_reference_list_with_existing_id(ObjRef* for_obj, ObjRef* ref_list)
{
    Block_Cache* cache;
    int i;

    assert(!is_void_reference(ref_list));

    cache = ref_list_cache;

    for (i = 0; i < REF_LIST_CACHE_SIZE; i++)
    {
	if (is_void_reference(&ref_list_cache[i].ref))
	{
	    cache = &ref_list_cache[i];
	    break;
	}

	if (ref_list_cache[i].age < cache->age)
	{
	    cache = &ref_list_cache[i];
	}
    }

    if (i == REF_LIST_CACHE_SIZE)    /* didn't find an empty one, write the oldest one out */
    {
	write_reference_list(cache);
    }

    copy_reference(&cache->ref, ref_list);

    ref_list_tick++;
    cache->age = ref_list_tick;

    cache->first_block = nwos_malloc(FILE_BLOCK_SIZE);

    assert(cache->first_block != NULL);

    memset(cache->first_block, 0, FILE_BLOCK_SIZE);  /* zero it out */

    cache->num_refs = 0;

    memset(cache->ivec, 0, sizeof(cache->ivec));

    nwos_fill_in_common_header(&cache->first_block->list.common_header, &cache->ref, &nwos_reference_list_class_ref);

    cache->first_block->list.common_header.flags = 0xffffffff;  /* mark it as dirty */

    copy_reference(&cache->first_block->list.common_header.id, &cache->ref);


#if 0
    printf("sizeof: %u chksum: %02x %02x %02x %02x\n", sizeof(ReferenceList),
	   rl.common_header.chksum[0], rl.common_header.chksum[1], rl.common_header.chksum[2], rl.common_header.chksum[3]);
#endif

    assert(cache->num_refs == 0);
    assert(cache->first_block != NULL);
    assert(is_void_reference(&cache->first_block->refs[cache->num_refs]));
}


void nwos_create_reference_list(ObjRef* for_obj, ObjRef* ref_list)
{
    nwos_generate_new_id(ref_list);

    nwos_create_reference_list_with_existing_id(for_obj, ref_list);
}


void nwos_add_to_references(ObjRef* ref, ObjRef* obj)
{
    EveryObject header;
    ObjRef class_ref;

    /* first verify we are didn't pass in a reference list as the object */
    /* I.E. meant to call add_to_reference_list instead. */

    nwos_get_object_class(obj, &class_ref);
    assert(!is_same_object(&class_ref, &nwos_reference_list_class_ref));

    nwos_read_object_headers_from_disk(obj, &header);

    nwos_add_to_reference_list(ref, &header.object.references);
}



/* Add a new reference to the reference list specified (ref_list is the id of the reference list itself). */

void nwos_add_to_reference_list(ObjRef* ref, ObjRef* ref_list)
{
    char log_msg[128];
    Block_Cache* cache;
    int running_total;
    int refs_remaining;
    Ref_List_Extra_Block* next_ptr;
    uint8 ivec[IVEC_SIZE];
    int seq;

#ifdef PUBLIC_MODE
    assert(nwos_reference_type(ref) == Public_Reference);
    assert(nwos_reference_type(ref_list) == Public_Reference);
#else
    assert(nwos_reference_type(ref) == Private_Reference);
    assert(nwos_reference_type(ref_list) == Private_Reference);
#endif

    cache = find_ref_list_in_cache(ref_list);

    /* first make sure it is a reference list */

    assert(is_same_object(&cache->first_block->list.common_header.class_definition, &nwos_reference_list_class_ref));

#ifdef LOG_REFERENCE_LIST
    snprintf(log_msg, sizeof(log_msg), "nwos_add_to_reference_list - object: %02x%02x%02x%02x - ref list: %02x%02x%02x%02x - num_refs: %d", 
	    ref->id[0], ref->id[1], ref->id[2], ref->id[3],
	    ref_list->id[0], ref_list->id[1], ref_list->id[2], ref_list->id[3],
	    cache->num_refs);
    nwos_log(log_msg);
#endif

    if (cache->num_refs <= MAX_REFS_IN_REF_LIST)    /* there is just one block */
    {
	assert(cache->first_block->next_block_ptr == NULL);         /* the next block pointer should be null */
	assert(is_void_reference(&cache->first_block->refs[cache->num_refs]));

	if (cache->num_refs < MAX_REFS_IN_REF_LIST)    /* there is room for more */
	{
	    memcpy(&cache->first_block->refs[cache->num_refs], ref, sizeof(ObjRef));
	    cache->num_refs++;
	    memset(&cache->first_block->refs[cache->num_refs], 0, sizeof(ObjRef));
	    cache->first_block->list.common_header.flags = 0xffffffff;     /* mark it as dirty */
	}
	else       /* no room for more, must expand to 2 blocks */
	{
	    assert(cache->num_refs == MAX_REFS_IN_REF_LIST);

	    nwos_generate_new_id(&cache->first_block->refs[MAX_REFS_IN_REF_LIST]);    /* id of new list */

	    memset(ivec, 0, sizeof(ivec));

	    cache->first_block->list.common_header.flags = 0;    /* remove dirty flag */

	    nwos_crc32_calculate((uint8*) &cache->first_block->refs, 
				 (cache->num_refs + 1) * sizeof(ObjRef),               /* include void ptr or next_block_ref */
				 cache->first_block->list.common_header.data_chksum);

	    nwos_write_object_to_disk_and_encrypt(&cache->ref, &cache->first_block->list, FILE_BLOCK_SIZE, cache->ivec, nwos_random_sequence[0]);

	    next_ptr = nwos_malloc(sizeof(Ref_List_Extra_Block));

	    assert(next_ptr != NULL);

	    cache->first_block->next_block_ptr = next_ptr;

	    next_ptr->next_block_ptr = NULL;

	    snprintf(log_msg, sizeof(log_msg), "add_reference - first_block ref: %02x%02x%02x%02x obj: %p  next_block ref: %02x%02x%02x%02x obj: %p",
		   cache->ref.id[0], cache->ref.id[2], cache->ref.id[2], cache->ref.id[3], cache->first_block,
		   cache->first_block->next_block_ref.id[0], cache->first_block->next_block_ref.id[1], 
		   cache->first_block->next_block_ref.id[2], cache->first_block->next_block_ref.id[3], 
		   cache->first_block->next_block_ptr);
	    nwos_log(log_msg);

	    next_ptr->dirty = -1;     /* mark it as dirty */
	    memcpy(&next_ptr->id, &cache->first_block->next_block_ref, sizeof(ObjRef));
	    memcpy(&next_ptr->refs[0], ref, sizeof(ObjRef));
	    next_ptr->refs[1].word = 0;              /* void reference to mark the end of the list */

	    cache->num_refs++;
	}
    }
    else
    {
	assert(cache->first_block->next_block_ptr != NULL);         /* the next block pointer should not be null */
	assert(sizeof(ReferenceList) + MAX_REFS_IN_REF_LIST * sizeof(ObjRef) == FILE_BLOCK_SIZE - 4);

	running_total = MAX_REFS_IN_REF_LIST;
	next_ptr = cache->first_block->next_block_ptr;        /* get pointer to first sidecar */

	assert(is_same_object(&cache->first_block->next_block_ref, &next_ptr->id));  /* next pointer ref matches */

	seq = 1;

	while (running_total < cache->num_refs)    /* read more blocks */
	{
	    refs_remaining = cache->num_refs - running_total;

	    if (refs_remaining <= MAX_REFS_IN_SIDECAR)
	    {
		assert(next_ptr->next_block_ptr == NULL);  /* the next block pointer should be null */
		assert(is_void_reference(&next_ptr->refs[refs_remaining]));

		if (refs_remaining < MAX_REFS_IN_SIDECAR)    /* there is room for more */
		{
		    memcpy(&next_ptr->refs[refs_remaining], ref, sizeof(ObjRef));
		    next_ptr->refs[refs_remaining + 1].word = 0;  /* set the next ref to null */
		    next_ptr->dirty = -1;                   /* mark it as dirty */
		}
		else       /* no room for more, must add new block */
		{
		    assert(refs_remaining == MAX_REFS_IN_SIDECAR);

		    nwos_generate_new_id(&next_ptr->next_block_ref);    /* id of new list */

		    next_ptr->dirty = 0;   /* remove dirty flag */

		    nwos_crc32_calculate((uint8*) &next_ptr->refs, 
					 (refs_remaining + 1) * sizeof(ObjRef),        /* include void ptr or next_block_ref */
					 next_ptr->checksum);

		    nwos_write_object_to_disk_and_encrypt(&next_ptr->id, &next_ptr->dirty, FILE_BLOCK_SIZE, cache->ivec, nwos_random_sequence[seq]);

		    next_ptr->next_block_ptr = nwos_malloc(sizeof(Ref_List_Extra_Block));

		    assert(next_ptr->next_block_ptr != NULL);

		    next_ptr->next_block_ptr->next_block_ptr = NULL;

	    snprintf(log_msg, sizeof(log_msg), "add_reference - next_block ref: %02x%02x%02x%02x obj: %p  next_block ref: %02x%02x%02x%02x obj: %p",
		   next_ptr->id.id[0], next_ptr->id.id[2], next_ptr->id.id[2], next_ptr->id.id[3], 
		   next_ptr,
		   next_ptr->next_block_ref.id[0], next_ptr->next_block_ref.id[1], 
		   next_ptr->next_block_ref.id[2], next_ptr->next_block_ref.id[3], 
		   next_ptr->next_block_ptr);
	    nwos_log(log_msg);

		    next_ptr->next_block_ptr->dirty = -1;     /* mark it as dirty */
		    memcpy(&next_ptr->next_block_ptr->id, &next_ptr->next_block_ref, sizeof(ObjRef));
		    memcpy(&next_ptr->next_block_ptr->refs[0], ref, sizeof(ObjRef));
		    next_ptr->next_block_ptr->refs[1].word = 0;              /* void reference to mark the end of the list */
		}

		running_total += refs_remaining;
	    }
	    else
	    {
		assert(next_ptr->next_block_ptr != NULL);  /* the next block pointer should not be null */
		running_total += MAX_REFS_IN_SIDECAR;

		/* next pointer ref matches */
		assert(is_same_object(&next_ptr->next_block_ref, &next_ptr->next_block_ptr->id));
	    }

	    next_ptr = next_ptr->next_block_ptr;

	    seq = (seq + 1) % NUM_STORED_SEQ;    /* wrap around if we've used all of them */
	}

	cache->num_refs++;

	assert(running_total + 1 == cache->num_refs);
    }
}


void nwos_remove_from_references(ObjRef* ref, ObjRef* obj)
{
    EveryObject header;
    ObjRef class_ref;

    /* first verify we are didn't pass in a reference list as the object */
    /* I.E. meant to call remove_from_reference_list instead. */

    nwos_get_object_class(obj, &class_ref);
    assert(!is_same_object(&class_ref, &nwos_reference_list_class_ref));

    nwos_read_object_headers_from_disk(obj, &header);

    nwos_remove_from_reference_list(ref, &header.object.references);
}



/* Add a new reference to the reference list specified (ref_list is the id of the reference list itself). */

void nwos_remove_from_reference_list(ObjRef* ref, ObjRef* ref_list)
{
    char log_msg[128];
    Block_Cache* cache;
    int i;

#ifdef PUBLIC_MODE
    assert(nwos_reference_type(ref) == Public_Reference);
    assert(nwos_reference_type(ref_list) == Public_Reference);
#else
    assert(nwos_reference_type(ref) == Private_Reference);
    assert(nwos_reference_type(ref_list) == Private_Reference);
#endif

    cache = find_ref_list_in_cache(ref_list);

    /* first make sure it is a reference list */

    assert(is_same_object(&cache->first_block->list.common_header.class_definition, &nwos_reference_list_class_ref));

    snprintf(log_msg, sizeof(log_msg), "nwos_remove_from_reference_list - object: %02x%02x%02x%02x - ref list: %02x%02x%02x%02x - num_refs: %d", 
	    ref->id[0], ref->id[1], ref->id[2], ref->id[3],
	    ref_list->id[0], ref_list->id[1], ref_list->id[2], ref_list->id[3],
	    cache->num_refs);
    nwos_log(log_msg);

    if (cache->num_refs <= MAX_REFS_IN_REF_LIST)    /* there is just one block */
    {
	assert(cache->first_block->next_block_ptr == NULL);         /* the next block pointer should be null */
	assert(is_void_reference(&cache->first_block->refs[cache->num_refs]));

	cache->num_refs--;

	if (!is_same_object(ref, &cache->first_block->refs[cache->num_refs]))    /* not the last one in the list */
	{
	    for (i = 0; i < cache->num_refs; i++)
	    {
		if (is_same_object(ref, &cache->first_block->refs[i])) break;
	    }
		
	    assert(i < cache->num_refs);    /* make sure it was in the list */

	    copy_reference(&cache->first_block->refs[i], &cache->first_block->refs[cache->num_refs]);
	}

	void_reference(&cache->first_block->refs[cache->num_refs]);
	cache->first_block->list.common_header.flags = 0xffffffff;     /* mark it as dirty */
    }
    else
    {
	bool this_version_cannot_handle_removing_from_long_reference_list = false;
	assert(this_version_cannot_handle_removing_from_long_reference_list);
    }
}


/* flush any reference lists that are dirty out to disk */

void nwos_flush_dirty_ref_lists()
{
    int i;

    for (i = 0; i < REF_LIST_CACHE_SIZE; i++)
    {
	if (!is_void_reference(&ref_list_cache[i].ref))
	{
	    write_reference_list(&ref_list_cache[i]);
	}
    }
}


ReferenceList* nwos_malloc_reference_list(ObjRef* ref)
{
    size_t ref_list_size;
    ReferenceList* ref_list;

    ref_list_size = nwos_reference_list_size(ref);

    ref_list = nwos_malloc(ref_list_size);

    if (ref_list == NULL) 
    {
	perror("reading reference list");
	exit(1);
    }

    nwos_read_reference_list_from_disk(ref, ref_list, ref_list_size);

    ref_list->common_header.num_refs = (ref_list_size - sizeof(CommonHeader)) / sizeof(ObjRef);

    return ref_list;
}


/* This creates a reference list of all objects of a class including both public and private */

ReferenceList* nwos_malloc_class_reference_list(char* class_name)
{
    size_t ref_list_size;
    ReferenceList* ref_list;
    ReferenceList* public_ref_list;
    ReferenceList* private_ref_list = NULL;
    C_struct_Class_Definition class_def;
    ObjRef public_ref;
    ObjRef private_ref;
    ObjRef obj_class_ref;
    int i;
    int num_refs = 0;

    assert(nwos_find_public_class_definition(class_name, &public_ref));
    nwos_read_class_definition(&public_ref, &class_def);
    public_ref_list = nwos_malloc_reference_list(&class_def.header.object.references);
    ref_list_size = public_ref_list->common_header.num_refs;

    if (nwos_find_private_class_definition(class_name, &private_ref))
    {
	nwos_read_class_definition(&private_ref, &class_def);
	private_ref_list = nwos_malloc_reference_list(&class_def.header.object.references);
	ref_list_size += private_ref_list->common_header.num_refs;
    }

    assert(ref_list_size > 0);

    ref_list = nwos_malloc(ref_list_size);

    if (ref_list == NULL) 
    {
	perror("reading reference list");
	exit(1);
    }

    if (private_ref_list != NULL)
    {
	for (i = 0; i < private_ref_list->common_header.num_refs; i++)
	{
	    nwos_get_object_class(&private_ref_list->references[i], &obj_class_ref);   /* find out what kind of object it is */

	    if (is_same_object(&obj_class_ref, &nwos_private_class_definition_class_ref))   /* it is a class definition object */
	    {
		copy_reference(&ref_list->references[num_refs], &private_ref_list->references[i]);
		num_refs++;
	    }
	}

	nwos_free_reference_list(private_ref_list);
	private_ref_list = NULL;
    }

    for (i = 0; i < public_ref_list->common_header.num_refs; i++)
    {
	nwos_get_object_class(&public_ref_list->references[i], &obj_class_ref);   /* find out what kind of object it is */

	if (is_same_object(&obj_class_ref, &nwos_public_class_definition_class_ref))   /* it is a class definition object */
	{
	    copy_reference(&ref_list->references[num_refs], &public_ref_list->references[i]);
	    num_refs++;
	}
    }

    nwos_free_reference_list(public_ref_list);
    public_ref_list = NULL;

    ref_list->common_header.num_refs = num_refs;

    return ref_list;
}


void nwos_free_reference_list(ReferenceList* ref_list)
{
    nwos_free(ref_list);
}

